/*
 * TuxNES
 *
 * Written by Ben Sittler
 *
 * Description: This file is largely responsible for mixing the waveform and
 * pumping it out the audio device.
 *
 * $Id: sound.c,v 1.14 2000/07/08 02:17:12 bsittler Exp $ */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <endian.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_LINUX_SOUNDCARD_H
#include <linux/soundcard.h>
#endif /* HAVE_LINUX_SOUNDCARD_H */

#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#endif /* HAVE_SYS_SOUNDCARD_H */

#include "consts.h"
#include "globals.h"
#include "mapper.h"
#include "sound.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

/* global sound parameters */
struct SoundConfig sound_config = {
     /* char *audiofile */
     0,
     /* int	audiorate */
     8000,
     /* int	audiostereo */
     0,
     /* float	max_sound_delay */
     0.0,
     /* int	reverb */
     0,
     /* int	use_sin */
     1
};

struct SoundData sound_data = {
     /* int	SQR1_FR2_write */
     0,
     /* int	SQR2_FR2_write */
     0,
     /* int	TRI_FR2_write */
     0,
     /* int	NOI_FR2_write */
     0
};

/* the currently selected sample format */
struct SampleFormat *sample_format = 0;

/* sample format numbers */
#ifndef AFMT_MU_LAW
#  define AFMT_MU_LAW 1
#endif
#ifndef AFMT_U8
#  define AFMT_U8 8
#endif
#ifndef AFMT_S16_LE
#  define AFMT_S16_LE 16
#endif
#ifndef AFMT_S16_BE
#  define AFMT_S16_BE 32
#endif
#ifndef AFMT_S8
#  define AFMT_S8 64
#endif
#ifndef AFMT_U16_LE
#  define AFMT_U16_LE 128
#endif
#ifndef AFMT_U16_BE
#  define AFMT_U16_BE 256
#endif

/* table of sample formats, terminated by { 0, 0, 0, 0 } */
struct SampleFormat sample_formats[] = {
     { "mu8",    "8-bit Mu-Law encoded",
       AFMT_MU_LAW },
     { "8",      "8-bit unsigned",
       AFMT_U8 },
     { "8s",     "8-bit signed",
       AFMT_S8 },
     { "16",     "16-bit signed",
#if __BYTE_ORDER == __BIG_ENDIAN
       AFMT_S16_BE
#else
       AFMT_S16_LE
#endif
     },
     { "16u",    "16-bit unsigned",
#if __BYTE_ORDER == __BIG_ENDIAN
       AFMT_U16_BE
#else
       AFMT_U16_LE
#endif
     },
     { "le16",   "16-bit signed (little-endian)",
       AFMT_S16_LE },
     { "le16u",  "16-bit unsigned (little-endian)",
       AFMT_U16_LE },
     { "be16",   "16-bit signed (big-endian)",
       AFMT_S16_BE },
     { "be16u",  "16-bit unsigned (big-endian)",
       AFMT_U16_BE },
     { 0, 0, 0 }
};

static int audiofd = -1;
static int sample_format_number;

/*
 * Sound Generator
 * This is a gruesome hack, and it sounds really bad. Sorry.
 */
#define wave1(angle,duty) \
        (((angle) < (2 * M_PI * (duty))) ? 0.707 : -0.707)
#define wave2(angle,duty) \
        sin(angle)
#define wave3(angle,duty) \
        wave2((angle < (2 * M_PI * duty)) ? (angle * 0.50 / duty) : \
                       ((angle - 2 * M_PI) * 0.50 / (1.0 - duty)), 0.5)
#define wave4(angle,duty) \
        ((angle<M_PI/2) ? angle : (angle<3*M_PI/2 ? M_PI-angle : angle-2*M_PI))

#if HAVE_LIBM
#define wave5 wave3
#else
#define wave5 wave1
#endif

#define square_wave(a,d) (sound_config.use_sin ? wave5((a),(d)) : wave1((a),(d)))

#define triangle_wave(angle) wave4(angle, 0.50)

/* NTSC frequency multiplier */
#define BASE_FREQ 111860.78
/* NTSC sound update rate (== vertical retrace rate) */
#define UPDATE_FREQ 60.0
/* PAL frequency multiplier */
/* #define BASE_FREQ 110840.47 */
/* PAL sound update rate (== vertical retrace rate) */
/* #define UPDATE_FREQ 50.0 */

int
InitAudio(int argc, char **argv)
{
  /* Open an audio stream */
  if (sound_config.audiofile)
    {
      if ((audiofd = open (sound_config.audiofile, O_CREAT | O_WRONLY | O_APPEND, 0666 & ~mask)) < 0)
        {
          perror (sound_config.audiofile);
	  return 1;
        }
      else
        {
#ifdef SNDCTL_DSP_RESET
	  if (! ioctl (audiofd, SNDCTL_DSP_RESET)) {
	    int desired_audiorate;
	    int desired_audiostereo;
	    struct SampleFormat *desired_sample_format = sample_format;

	    desired_audiorate = sound_config.audiorate;
	    if (ioctl (audiofd, SNDCTL_DSP_SPEED, &sound_config.audiorate))
	      perror (sound_config.audiofile);
	    if (sound_config.audiorate != desired_audiorate)
	      {
		fprintf (stderr, "%s: wanted %d Hz, got %d Hz\n",
			 sound_config.audiofile,
			 desired_audiorate,
			 sound_config.audiorate);
	      }
	    desired_audiostereo = sound_config.audiostereo;
	    if (ioctl (audiofd, SNDCTL_DSP_STEREO, &sound_config.audiostereo))
	      perror (sound_config.audiofile);
	    if (sound_config.audiostereo != desired_audiostereo)
	      fprintf (stderr, "%s: wanted %s, got %s (expect failure)\n", sound_config.audiofile,
		       desired_audiostereo ? "stereo" : "monaural",
		       sound_config.audiostereo ? "stereo" : "monaural");
	    sample_format_number = sample_format->number;
	    if (ioctl (audiofd, SNDCTL_DSP_SAMPLESIZE, &sample_format_number))
	      perror (sound_config.audiofile);
	    if (sample_format_number != sample_format->number)
	      {
		for (sample_format = sample_formats; sample_format->name; sample_format ++)
	          {
		    if (sample_format->number == sample_format_number) break;
		  }
		fprintf (stderr, "%s: wanted %s samples [%d], got %s samples [%d]\n",
			 sound_config.audiofile,
			 desired_sample_format->fullname
			 ? desired_sample_format->fullname
			 : "(unknown sample format)",
			 desired_sample_format->number,
			 sample_format->fullname
			 ? sample_format->fullname
			 : "(unknown sample format)",
			 sample_format_number);
	      }
	  }
#endif /* SNDCTL_DSP_RESET */
	  if (verbose)
	    fprintf (stderr,
		     "Appending %d Hz %s samples to %s\n",
		     sound_config.audiorate,
		     sample_format->fullname
		     ? sample_format->fullname
		     : "(unknown sample format)",
		     sound_config.audiofile);
        }
    }
  return 0;
}

void
UpdateAudio(void)
{
  static unsigned char *buf = 0; /* dynamically allocated on first call */
  int nsamples;                  /* flexible, size dithers */
  static double samerr = 0.0;   /* accumulated error in buffer size */
  unsigned i;
  /* for reverb */
  static int *rvbuf = 0;
  static unsigned rvbufsiz = 0;
  static unsigned rvoff = 0;

  unsigned SCR = RAM[0x4015];

  unsigned SQR1_enabled = SCR & 0x01;
  unsigned SQR2_enabled = SCR & 0x02;
  unsigned TRI_enabled = SCR & 0x04;
  unsigned NOI_enabled = SCR & 0x08;
  unsigned PCM_enabled = SCR & 0x10;

  /* duty cycle decoding table */
  static double duty[4] =
  {0.875,
   0.750,
   0.500,
   0.250};

  /* SQR1: Square Wave Channel #1 */
  unsigned SQR1_CR1 = RAM[0x4000];      /* Control Register 1 */
  unsigned SQR1_CR2 = RAM[0x4001];      /* Control Register 2 */
  unsigned SQR1_FR1 = RAM[0x4002];      /* Frequency (Inverse) Register 1 */
  unsigned SQR1_FR2 = RAM[0x4003];      /* Frequency (Inverse) Register 2 */
  static unsigned SQR1_FR2_cache = 0;

  double SQR1_duty = duty[SQR1_CR1 >> 6];       /* Duty Cycle */
  unsigned SQR1_hold = (SQR1_CR1 >> 5) & 0x01;  /* Loop note */
  unsigned SQR1_fixe = (SQR1_CR1 >> 4) & 0x01;  /* Fixed envelope */
  unsigned SQR1_rate = SQR1_CR1 & 0x0f;         /* volume / Decay speed */
  unsigned SQR1_vfrq = SQR1_CR2 >> 7;   /* Variable frequency */
  unsigned SQR1_vfsp = (SQR1_CR2 >> 4) & 0x07;  /* Frequency change speed */
  unsigned SQR1_lowr = (SQR1_CR2 >> 3) & 0x01;  /* Lower frequency */
  unsigned SQR1_frng = SQR1_CR2 & 0x07;         /* Frequency range */
  unsigned SQR1_peri = SQR1_FR1 | ((SQR1_FR2 << 8) & 0x0700);   /* Period */
  unsigned SQR1_time = SQR1_FR2 >> 3;   /* Note timer */

  double SQR1_freq = BASE_FREQ / (SQR1_peri + 1);   /* frequency */
  static double SQR1_angl = 0.0;        /* phase angle */
  static int SQR1_timer = 0;    /* timer */
  static int SQR1_enve = 15; /* envelope */

  /* SQR2: Square Wave Channel #1 */
  unsigned SQR2_CR1 = RAM[0x4004];      /* Control Register 1 */
  unsigned SQR2_CR2 = RAM[0x4005];      /* Control Register 2 */
  unsigned SQR2_FR1 = RAM[0x4006];      /* Frequency (Inverse) Register 1 */
  unsigned SQR2_FR2 = RAM[0x4007];      /* Frequency (Inverse) Register 2 */
  static unsigned SQR2_FR2_cache = 0;

  double SQR2_duty = duty[SQR2_CR1 >> 6];       /* Duty Cycle */
  unsigned SQR2_hold = (SQR2_CR1 >> 5) & 0x01;  /* Loop note */
  unsigned SQR2_fixe = (SQR2_CR1 >> 4) & 0x01;  /* Fixed envelope */
  unsigned SQR2_rate = SQR2_CR1 & 0x0f;         /* volume / Decay speed */
  unsigned SQR2_vfrq = SQR2_CR2 >> 7;   /* Variable frequency */
  unsigned SQR2_vfsp = (SQR2_CR2 >> 4) & 0x07;  /* Frequency change speed */
  unsigned SQR2_lowr = (SQR2_CR2 >> 3) & 0x01;  /* Lower frequency */
  unsigned SQR2_frng = SQR2_CR2 & 0x07;         /* Frequency range */
  unsigned SQR2_peri = SQR2_FR1 | ((SQR2_FR2 << 8) & 0x0700);   /* Period */
  unsigned SQR2_time = SQR2_FR2 >> 3;   /* Note timer */

  double SQR2_freq = BASE_FREQ / (SQR2_peri + 1);   /* frequency */
  static double SQR2_angl = 0.0;        /* phase angle */
  static int SQR2_timer = 0;    /* timer */
  static int SQR2_enve = 15; /* envelope */

  /* TRI: Triangle Wave Channel */
  unsigned TRI_CR = RAM[0x4008];        /* Control Register */
  unsigned TRI_FR1 = RAM[0x400A];       /* Frequency (Inverse) Register 1 */
  unsigned TRI_FR2 = RAM[0x400B];       /* Frequency (Inverse) Register 2 */
  static unsigned TRI_FR2_cache = 0;

  unsigned TRI_hold = TRI_CR >> 7;    /* Loop note */
  unsigned TRI_tim2 = TRI_CR & 0x7f;   /* Note timer 2 */
  unsigned TRI_peri = TRI_FR1 | ((TRI_FR2 << 8) & 0x0700);      /* Period */
  unsigned TRI_time = TRI_FR2 >> 3;     /* Note timer */

  double TRI_freq = 0.5 * BASE_FREQ / (TRI_peri + 1);        /* frequency */
  static double TRI_angl = 0.0; /* phase angle */
  static int TRI_timer = 0;     /* timer */
  static int TRI_hold2 = 0;     /* note timer 2 hold */
  static int TRI_timer2 = 0;    /* note timer 2 */

  /* NOI: Noise Channel */
  unsigned NOI_CR = RAM[0x400C];        /* Control Register */
  unsigned NOI_FR1 = RAM[0x400E];       /* Frequency (Inverse) Register 1 */
  unsigned NOI_FR2 = RAM[0x400F];       /* Frequency (Inverse) Register 2 */
  static unsigned NOI_FR2_cache = 0;

  unsigned NOI_hold = (NOI_CR >> 5) & 0x01;    /* Loop note */
  unsigned NOI_fixe = (NOI_CR >> 4) & 0x01;    /* Fixed envelope */
  unsigned NOI_rate = NOI_CR & 0x0f;   /* volume / Decay speed */
  unsigned NOI_type = NOI_FR1 >> 7;    /* Noise type */
  unsigned NOI_peri = NOI_FR1 & 0x0f;  /* Period */
  unsigned NOI_time = NOI_FR2 >> 3;     /* Note timer */

  static int NOI_periods[16] = {
    128, 256, 512, 1024, 2048, 3069, 4096, 5113,
    6461, 8192, 12175, 16384, 23243, 32768, 65536, 131072
  };

  double NOI_freq = 16 * BASE_FREQ / NOI_periods[NOI_peri];
  static double NOI_angl = 0.0; /* phase angle */
  static int NOI_timer = 0;     /* timer */
  static int NOI_enve = 15; /* envelope */
  static int bytes_per_sample = 0;

  /* if sound is disabled, we stop right here */
  if (audiofd < 0)
    return;

  /* calculate bytes-per-sample */
  if (! bytes_per_sample)
    switch (sample_format_number)
      {
      case AFMT_MU_LAW:
      case AFMT_U8:
      case AFMT_S8:
	bytes_per_sample = 1;
	break;
      case AFMT_S16_LE:
      case AFMT_S16_BE:
      case AFMT_U16_LE:
      case AFMT_U16_BE:
	bytes_per_sample = 2;
	break;
      default:
	fprintf (stderr, "%s [%d]: not a supported sample format\n",
		 sample_format->fullname
		 ? sample_format->fullname
		 : "(unknown sample format)",
		 sample_format_number);
	audiofd = -1;
	return;
      }
  /* allocate the buffer, if we haven't already */
  if ((! buf) && !(buf = (unsigned char *)
		   malloc ((size_t)(2 * sound_config.audiorate * bytes_per_sample
				    / UPDATE_FREQ + 1))))
    {
      perror ("can't allocate sound buffer (sound disabled)");
      close (audiofd);
      audiofd = -1;
      return;
    }
  /* reverb initialization */
  if (sound_config.reverb && !rvbuf)
    {
      /* calculate reverb buffer size */
      if (!(rvbufsiz = sound_config.audiorate / 10))
	{
	  rvbufsiz = 1;
	}
      /* allocate the reverb buffer */
      if (!(rvbuf = (int *) malloc (rvbufsiz * sizeof(int))))
	{
	  perror ("can't allocate reverb buffer (reverb disabled)");
	  sound_config.reverb = 0;
	}
      else
	{
	  int ri;

	  for (ri = 0; ri < rvbufsiz; ri ++)
	    rvbuf[ri] = 0;
	}
    }

  /* distribute the fractional buffer size */
  nsamples = (int)(((halfspeed ? 2 : 1) * sound_config.audiorate) /
		   (UPDATE_FREQ * (doublespeed ? doublespeed : 1)));
  samerr += (((halfspeed ? 2 : 1) * sound_config.audiorate) /
	     (UPDATE_FREQ * (doublespeed ? doublespeed : 1))) - nsamples;
  if (samerr >= 1.0) {
    nsamples ++;
    samerr -= 1.0;
  }

  /* handle note timers */

  if (SQR1_enabled) {
    if (SQR1_timer && !SQR1_hold)
      SQR1_timer--;
    if ((SQR1_FR2 != SQR1_FR2_cache) || sound_data.SQR1_FR2_write)
      {
	if (verbose && !sound_data.SQR1_FR2_write)
	  fprintf(stderr, "detected unhandled SQR1 write\n");
	SQR1_angl = 0.0;
	SQR1_timer = SQR1_time << 2;
	SQR1_enve = 15;
	SQR1_FR2_cache = SQR1_FR2;
	sound_data.SQR1_FR2_write = 0;
      }
    if (SQR1_fixe)
      SQR1_enve = SQR1_rate;
    else if (SQR1_enve || SQR1_hold)
      SQR1_enve = (SQR1_enve - 1) & 0xf;
  } else {
    SQR1_timer = 0;
    SQR1_FR2_cache = SQR1_FR2;
  }

  if (SQR2_enabled) {
    if (SQR2_timer && !SQR2_hold)
      SQR2_timer--;
    if ((SQR2_FR2 != SQR2_FR2_cache) || sound_data.SQR2_FR2_write)
      {
	if (verbose && !sound_data.SQR2_FR2_write)
	  fprintf(stderr, "detected unhandled SQR2 write\n");
	SQR2_angl = 0.0;
	SQR2_timer = SQR2_time << 2;
	SQR2_enve = 15;
	SQR2_FR2_cache = SQR2_FR2;
	sound_data.SQR2_FR2_write = 0;
      }
    if (SQR2_fixe)
      SQR2_enve = SQR2_rate;
    else if (SQR2_enve || SQR2_hold)
      SQR2_enve = (SQR2_enve - 1) & 0xf;
  } else {
    SQR2_timer = 0;
    SQR2_FR2_cache = SQR2_FR2;
  }

  if (TRI_enabled) {
    if (TRI_timer && !TRI_hold)
      TRI_timer--;
    if (TRI_timer2 && !TRI_hold2)
      TRI_timer2 --;
    if ((TRI_FR2 != TRI_FR2_cache) || sound_data.TRI_FR2_write)
      {
	if (verbose && !sound_data.TRI_FR2_write)
	  fprintf(stderr, "detected unhandled TRI write\n");
	TRI_angl = 0.0;
	TRI_timer = TRI_time;
	TRI_hold2 = TRI_hold;
	TRI_timer2 = (TRI_tim2 + 1) >> 2;
	TRI_FR2_cache = TRI_FR2;
	sound_data.TRI_FR2_write = 0;
      }
    if (TRI_hold2 && !TRI_hold)
      TRI_hold2 = 0;
    if (TRI_hold2)
      TRI_timer2 = (TRI_tim2 + 1) >> 2;
  } else {
    TRI_timer = 0;
    TRI_FR2_cache = TRI_FR2;
  }

  if (NOI_enabled) {
    if (NOI_timer && !NOI_hold)
      NOI_timer--;
    if ((NOI_FR2 != NOI_FR2_cache) || sound_data.NOI_FR2_write)
      {
	if (verbose && !sound_data.NOI_FR2_write)
	  fprintf(stderr, "detected unhandled NOI write\n");
	NOI_angl = 0.0;
	NOI_timer = NOI_time;
	NOI_enve = 15;
	NOI_FR2_cache = NOI_FR2;
	sound_data.NOI_FR2_write = 0;
      }
    if (NOI_fixe)
      NOI_enve = NOI_rate;
    else if (NOI_enve || NOI_hold)
      NOI_enve = (NOI_enve - 1) & 0xf;
  } else {
    NOI_timer = 0;
    NOI_FR2_cache = NOI_FR2;
  }

  /* fill the buffer */
  for (i = 0; i < nsamples; i++)
    {
      double sqr1 = 0.0, sqr2 = 0.0, tri = 0.0, noi = 0.0;
      static int noi0 = 0;
      int sample = 0;

      if (SQR1_enabled && SQR1_timer && SQR1_enve && SQR1_peri
	  && !((SQR1_FR2 & 0x04) && SQR1_lowr))
	sqr1 = square_wave (SQR1_angl, SQR1_duty) * SQR1_enve / 15;
      SQR1_angl += 2 * M_PI * SQR1_freq / sound_config.audiorate;
      while (SQR1_angl >= 2 * M_PI)
        {
          SQR1_angl -= 2 * M_PI;
        }

      if (SQR2_enabled && SQR2_timer && SQR2_enve && SQR2_peri
	  && !((SQR2_FR2 & 0x04) && SQR2_lowr))
	sqr2 = square_wave (SQR2_angl, SQR2_duty) * SQR2_enve / 15;
      SQR2_angl += 2 * M_PI * SQR2_freq / sound_config.audiorate;
      while (SQR2_angl > 2 * M_PI)
        {
          SQR2_angl -= 2 * M_PI;
        }

      if (TRI_enabled && TRI_timer && TRI_timer2 && TRI_peri)
	tri = triangle_wave (TRI_angl);
      TRI_angl += 2 * M_PI * TRI_freq / sound_config.audiorate;
      while (TRI_angl > 2 * M_PI)
        {
          TRI_angl -= 2 * M_PI;
        }

      if (NOI_enabled && NOI_timer && NOI_enve && NOI_peri)
	noi = noi0 * square_wave (NOI_angl, 0.50) * NOI_enve / 15;
      NOI_angl += 2 * M_PI * NOI_freq / sound_config.audiorate;
      while (NOI_angl > 2 * M_PI)
        {
	  noi0 = (rand() >> 8) & 1;
          NOI_angl -= 2 * M_PI;
        }

      if (sound_config.reverb)
	sample = rvbuf[rvoff] >> 2;
      switch (sample_format_number)
	{
	case AFMT_MU_LAW:
	  {
	    sample +=
	      1024 * sqr1 + 1024 * sqr2 + 1024 * tri + 1024 * noi;
	    buf[i] = ulaw [sample & 0x3fff];
	    break;
	  }
	case AFMT_U8:
	  {
	    sample +=
	      128 + 16 * sqr1 + 16 * sqr2 + 16 * tri + 16 * noi;
	    buf[i] = sample;
	    break;
	  }
	case AFMT_S8:
	  {
	    sample +=
	      16 * sqr1 + 16 * sqr2 + 16 * tri + 16 * noi;
	    buf[i] = (unsigned char) sample;
	    break;
	  }
	case AFMT_S16_LE:
	  {
	    sample +=
	      4096 * sqr1 + 4096 * sqr2 + 4096 * tri + 4096 * noi;
	    buf[i*2] = sample & 0xff;
	    buf[i*2 + 1] = (sample & 0xff00) >> 8;
	    break;
	  }
	case AFMT_S16_BE:
	  {
	    sample +=
	      4096 * sqr1 + 4096 * sqr2 + 4096 * tri + 4096 * noi;
	    buf[i*2 + 1] = sample & 0xff;
	    buf[i*2] = (sample & 0xff00) >> 8;
	    break;
	  }
	case AFMT_U16_LE:
	  {
	    sample +=
	      32768 + 4096 * sqr1 + 4096 * sqr2 + 4096 * tri + 4096 * noi;
	    buf[i*2] = sample & 0xff;
	    buf[i*2 + 1] = (sample & 0xff00) >> 8;
	    break;
	  }
	case AFMT_U16_BE:
	  {
	    sample +=
	      32768 + 4096 * sqr1 + 4096 * sqr2 + 4096 * tri + 4096 * noi;
	    buf[i*2 + 1] = sample & 0xff;
	    buf[i*2] = (sample & 0xff00) >> 8;
	    break;
	  }
	}
      if (sound_config.reverb)
	{
	  rvbuf[rvoff] = sample;
	  rvoff = (rvoff + 1) % rvbufsiz;
	}
    }

  /* play it! */
  write (audiofd, buf, nsamples * bytes_per_sample);
#ifdef SNDCTL_DSP_GETODELAY
  if (sound_config.max_sound_delay > 0.0) {
    int odelay;
    static int maxdelay = 0;
    
    if (! ioctl (audiofd, SNDCTL_DSP_GETODELAY, &odelay))
      {
	if (odelay > maxdelay)
	  {
	    maxdelay = odelay;
	  }
	if (odelay * 1.0 / (sound_config.audiorate * bytes_per_sample) >= sound_config.max_sound_delay)
	  {
	    if (! ioctl (audiofd, SNDCTL_DSP_SYNC))
	      {
		desync = 1;
  		if (verbose)
  		  {
  		    fprintf (stderr, "Warning: %.3f sec sound delay, resynchronizing\n",
			     odelay * 1.0 / (sound_config.audiorate * bytes_per_sample));
  		  }
	      }
	  }
      }
  }
#endif /* SNDCTL_DSP_GETODELAY */
}
